@@ -797,7 +797,7 @@ var NetworkAgent = class {
         }
 
         const prop = plugin.lookup_property('GNOME', 'supports-external-ui-mode');
-        const trimmedProp = prop ? prop.trim().toLowerCase() : '';
+        const trimmedProp = prop?.trim().toLowerCase() ?? '';
 
         return {
             fileName,
