@@ -57,11 +57,11 @@ var RadialShaderEffect = GObject.registerClass({
     }
 
     set brightness(v) {
-        if (this._brightness == v)
+        if (this._brightness === v)
             return;
         this._brightness = v;
         this.set_uniform_float(this._brightnessLocation,
-                               1, [this._brightness]);
+            1, [this._brightness]);
         this.notify('brightness');
     }
 
@@ -70,11 +70,11 @@ var RadialShaderEffect = GObject.registerClass({
     }
 
     set sharpness(v) {
-        if (this._sharpness == v)
+        if (this._sharpness === v)
             return;
         this._sharpness = v;
         this.set_uniform_float(this._sharpnessLocation,
-                               1, [this._sharpness]);
+            1, [this._sharpness]);
         this.notify('sharpness');
     }
 });
